/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

//* TODO UV-based option

#define FE_FCO_DEBUG		FALSE
#define FE_FCO_OUTPUT_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void FollicleOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	//* page Edit
	catalog<Real>("pickRadius")=0.01;
	catalog<Real>("pickRadius","high")=1.0;
	catalog<String>("pickRadius","label")="Pick Radius";
	catalog<String>("pickRadius","page")="Edit";
	catalog<String>("pickRadius","hint")="Inner region of area select.";

	catalog<Real>("falloff")=0.0;
	catalog<String>("falloff","label")="Falloff";
	catalog<String>("falloff","page")="Edit";
	catalog<String>("falloff","hint")="Falloff extension of area select.";

	catalog<Real>("fuzzy")=0.0;
	catalog<Real>("fuzzy","max")=1.0;
	catalog<String>("fuzzy","label")="Fuzzy Weights";
	catalog<String>("fuzzy","page")="Edit";
	catalog<String>("fuzzy","hint")="Selection randomness.";

	//* page Data
	catalog<bool>("useUV")=false;
	catalog<String>("useUV","label")="Use UV";
	catalog<String>("useUV","page")="Data";
	catalog<String>("useUV","hint")=
			"Make changes using UV coordinates instead of"
			"simply snapping to the surface in world space.";

	catalog<String>("hideGroup")="hide";
	catalog<String>("hideGroup","label")="Hide Group";
	catalog<String>("hideGroup","page")="Data";
	catalog<String>("hideGroup","hint")=
			"Put all hidden points in this new group.";

	//* page Display
	catalog<Real>("singleRadius")=0.1;
	catalog<String>("singleRadius","label")="Single Radius";
	catalog<String>("singleRadius","page")="Display";
	catalog<String>("singleRadius","hint")=
			"Drawn radius during single select.";

	catalog<I32>("lineWidth")=2;
	catalog<I32>("lineWidth","min")=1;
	catalog<I32>("lineWidth","max")=10;
	catalog<String>("lineWidth","label")="Line Width";
	catalog<String>("lineWidth","page")="Display";
	catalog<String>("lineWidth","hint")="Thickness of box lines.";

	catalog<I32>("rimSize")=4;
	catalog<I32>("rimSize","min")=2;
	catalog<I32>("rimSize","max")=10;
	catalog<String>("rimSize","label")="Rim Size";
	catalog<String>("rimSize","page")="Display";
	catalog<String>("rimSize","hint")="Width of boxes.";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")="Enter toggles area select.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
//	catalog<bool>("Output Surface","recycle")=true;

	m_spWireframe=new DrawMode();
	m_spWireframe->setDrawStyle(DrawMode::e_wireframe);
	m_spWireframe->setPointSize(3.0);
	m_spWireframe->setLineWidth(2.0);

	m_spSolid=new DrawMode();
	m_spSolid->setDrawStyle(DrawMode::e_solid);
	m_spSolid->setBackfaceCulling(FALSE);
}

void FollicleOp::handle(Record& a_rSignal)
{
	const Real singleRadius=catalog<Real>("singleRadius");
	const Real pickRadius=catalog<Real>("pickRadius");
	const Real falloff=catalog<Real>("falloff");
	const Real fuzzy=catalog<Real>("fuzzy");

	const I32 lineWidth=catalog<I32>("lineWidth");
	const I32 rimSize=catalog<I32>("rimSize");

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

#if FE_FCO_DEBUG
	feLog("FollicleOp::handle brush %d brushed %d\n",
			spDrawBrush.isValid(),m_brushed);
#endif

	if(spDrawBrush.isValid())
	{
		m_spWireframe->setLineWidth(lineWidth);

		spDrawBrush->pushDrawMode(m_spWireframe);
		spDrawOverlay->pushDrawMode(m_spWireframe);

		m_brushed=TRUE;

		const Color black(0.0,0.0,0.0,1.0);
		const Color white(1.0,1.0,0.8,1.0);
		const Color grey(0.8,0.8,1.0,1.0);
		const Color red(0.8,0.0,0.0,1.0);
		const Color green(0.0,0.5,0.0,1.0);
		const Color blue(0.0,0.0,1.0,1.0);
		const Color cyan(0.0,0.8,1.0,1.0);
		const Color magenta(0.8,0.0,0.8,1.0);
		const Color yellow(1.0,1.0,0.0,1.0);
		const Color orange(1.0,0.5,0.0,1.0);
		const Color blank(0.0,0.0,0.0,0.0);

		const SpatialVector up(0.0,1.0,0.0);

		const Real sphereRadius=0.02;
		const SpatialVector sphereScale(sphereRadius,sphereRadius,sphereRadius);

		/***
			select/move			Left, Drag
			unselect			click off object
			add					middle
			hide selection		Del
			restore				ctrl-Del

			multi-select		ctrl-single, shift-area + drag
			soft-select			Falloff (dot aligned)
			area-select			Radius (dot aligned)
			fuzzy-select		Noise-Weighting

			Enter toggles single/area mode
			soft available in area mode
			fuzzy is parameter of soft
			weight color like maya soft select (yellow->red?)

			inner box - selected (color by weight)
			outline box - to select
		*/

		if(m_spDriver.isValid())
		{
			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			m_event.bind(windowEvent(a_rSignal));

#if FE_FCO_DEBUG
			feLog("%s\n",c_print(m_event));
#endif

			SpatialVector intersection;
			set(intersection);
			SpatialVector hitNormal;
			set(hitNormal);
			BWORD impacted=FALSE;

			const Real maxDistance=100.0;
			sp<SurfaceI::ImpactI> spImpact=m_spDriver->rayImpact(
					rRayOrigin,rRayDirection,maxDistance);
			if(spImpact.isValid())
			{
				sp<SurfaceTriangles::Impact> spTriImpact=spImpact;
				if(spTriImpact.isValid())
				{
					intersection=spTriImpact->intersection();
					hitNormal=spTriImpact->normal();
#if FE_FCO_DEBUG
					const I32 face=spTriImpact->face();
					const I32 pickIndex=spTriImpact->pickIndex();
					feLog("FollicleOp::handle face %d pick %d %s\n",
							face,pickIndex,c_print(intersection));
#endif

					impacted=TRUE;
				}
			}

			sp<ViewI> spView=spDrawOverlay->view();
			sp<ViewMode> spViewMode=spView->viewMode();
			SpatialTransform camera=spViewMode->camera();
			SpatialTransform invCamera;
			invert(invCamera,camera);

//			feLog("cam\n%s\n",c_print(camera));
//			feLog("inv\n%s\n",c_print(invCamera));

			const SpatialVector deltaX=invCamera.column(0);
			const SpatialVector deltaY=invCamera.column(1);

			I32 closest= -1;
			Real closestDist= -1.0;

			const U32 outputCount=m_spOutputPoint->count();
			m_colorRim.resize(outputCount);
			m_colorCenter.resize(outputCount);

			const Real extent=pickRadius+falloff;
			const Real rad2=extent*extent;
			for(I32 index=0;index<outputCount;index++)
			{
				m_colorCenter[index]=blank;

				const SpatialVector normal=
						m_spOutputNormal->spatialVector(index);
				const Real hitDot=dot(normal,invCamera.column(2));
				if(hitDot<0)
				{
					m_colorRim[index]=blank;
					continue;
				}

				const SpatialVector point=
						m_spOutputPoint->spatialVector(index);

				m_colorRim[index]=green;

				if(spImpact.isValid())
				{
					const Real dist2=magnitudeSquared(point-intersection);
					if(dist2<rad2)
					{
						const Real dist=sqrt(dist2);

						if(m_areaSelect)
						{
							Real weight=1.0;
							if(falloff>0.0 && dist>pickRadius)
							{
								weight-=(dist-pickRadius)/falloff;
							}
							if(fuzzy)
							{
								weight-=randomReal(0.0,fuzzy);
							}

							m_colorRim[index]=m_areaSelect?
									Color(0.0,0.0,weight,1.0): blue;
						}

						//* hidden ignored unless ctrl and shift both held
						U32 item=m_event.item();
						if((closest<0 || closestDist>dist) &&
							((item&WindowEvent::e_keyShift &&
							item&WindowEvent::e_keyControl) ||
							!hidden(index)))
						{
							closest=index;
							closestDist=dist;
						}
					}
				}
			}
			if(!m_areaSelect && closest>=0)
			{
				m_colorRim[closest]=blue;
			}

#if FALSE
			spDrawBrush->pushDrawMode(m_spSolid);
			spDrawBrush->draw(m_spDriver,&white);
			spDrawBrush->popDrawMode();
#endif

			if(m_event.isKeyPress(WindowEvent::e_itemAny))
			{
				U32 item=m_event.item();
				BWORD shifted=FALSE;
				if(item&WindowEvent::e_keyShift)
				{
					shifted=TRUE;
					item^=WindowEvent::e_keyShift;
				}
				BWORD ctrled=FALSE;
				if(item&WindowEvent::e_keyControl)
				{
					ctrled=TRUE;
					item^=WindowEvent::e_keyControl;
				}

				if(item==WindowEvent::e_keyCarriageReturn)
				{
					m_areaSelect=!m_areaSelect;
				}

				if(item==WindowEvent::e_keyDelete)
				{
					const I32 pickCount=m_pickArray.size();
					for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
					{
						const Pick& rPick=m_pickArray[pickIndex];
						const I32 pointIndex=rPick.m_index;

						hideToggle(pointIndex);
					}
					m_pickArray.clear();
				}
			}

			if(!m_dragging && m_event.isMouseRelease(WindowEvent::e_itemLeft))
			{
				U32 item=m_event.item();
				BWORD shifted=FALSE;
				if(item&WindowEvent::e_keyShift)
				{
					shifted=TRUE;
					item^=WindowEvent::e_keyShift;
				}
				BWORD ctrled=FALSE;
				if(item&WindowEvent::e_keyControl)
				{
					ctrled=TRUE;
					item^=WindowEvent::e_keyControl;
				}

				if(closest>=0)
				{
					if(ctrled)
					{
						//* NOTE can ctrl-shift-pick hidden
						if(shifted || !hidden(closest))
						{
							pickToggle(closest);
						}
					}
					else if(picked(closest)<1.0)
					{
						if(!shifted)
						{
							m_pickArray.clear();
						}
						if(m_areaSelect)
						{
							for(I32 index=0;index<outputCount;index++)
							{
								//* HACK using color for region check
								if(m_colorRim[index][2]>0.0 && !hidden(index))
								{
									pick(index,m_colorRim[index][2]);
								}
							}
						}
						if(!hidden(closest))
						{
							pick(closest);
						}
					}
				}
				else if(!ctrled)
				{
					m_pickArray.clear();
				}
			}

			const WindowEvent::MouseButtons buttons=m_event.mouseButtons();
			if(buttons)
			{
				//* TODO screen aware
				const Real moveScale=3e-3;

				//* TODO use last UV to current UV
				const Real uvScale=0.3;

				const I32 mouseX=m_event.mouseX();
				const I32 mouseY=m_event.mouseY();

				const I32 moveX=mouseX-m_lastX;
				const I32 moveY=mouseY-m_lastY;

				const Vector2 uv=spImpact.isValid()?
						spImpact->uv(): Vector2(0.0,0.0);

#if FALSE
				const SpatialVector deltaP=
						moveScale*(deltaX*moveX+deltaY*moveY);

				const SpatialVector du=spImpact->du();
				const SpatialVector dv=spImpact->dv();
				const SpatialVector delta=catalog<bool>("useUV")?
						uvScale*SpatialVector(dot(deltaP,du),dot(deltaP,dv)):
						deltaP;
#else
				const SpatialVector delta=catalog<bool>("useUV")?
						SpatialVector(uv-m_lastUV):
						moveScale*(deltaX*moveX+deltaY*moveY);
#endif

				if(buttons&WindowEvent::e_mbLeft)
				{
					if(m_event.isMouseDrag())
					{
						if(closest>=0)
						{
							movePicked(delta);
						}
					}
				}

				if(buttons&WindowEvent::e_mbMiddle)
				{
					if(impacted && m_event.isMousePress())
					{
						addFollicle(intersection);
					}
				}

				m_lastX=mouseX;
				m_lastY=mouseY;
				m_lastUV=uv;
				m_dragging=m_event.isMouseDrag();
			}

			if(impacted)
			{
				const I32 depth=0;
				drawPicker(spDrawBrush,intersection,up,hitNormal,
						m_areaSelect? pickRadius: singleRadius,
						depth,&yellow);
				if(m_areaSelect)
				{
					drawPicker(spDrawBrush,intersection,up,hitNormal,
							falloff>0.0? pickRadius+falloff: 1.05*pickRadius,
							depth,
							falloff>0.0? &orange: &black);
				}
			}
		}

		const I32 outputCount=m_spOutputPoint->count();
		const I32 pickCount=m_pickArray.size();
		for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
		{
			const Pick& rPick=m_pickArray[pickIndex];
			const I32 pointIndex=rPick.m_index;
			const Real pointWeight=rPick.m_weight;
			if(pointIndex>=0 && pointIndex<outputCount)
			{
//				m_colorRim[pointIndex]=yellow;
				m_colorCenter[pointIndex]=Color(1.0,pointWeight,0.0,1.0);
			}
		}

		sp<RecordArray> spRA=m_spState->getArray("Hide");
		if(spRA.isValid())
		{
			for(int i=0;i<spRA->length();i++)
			{
				const I32 pointIndex=m_aIndex(spRA,i);
				if(pointIndex>=0 && pointIndex<outputCount)
				{
					m_colorRim[pointIndex]=red;
				}
			}
		}

		for(I32 index=0;index<outputCount;index++)
		{
			const Color& rColorCenter=m_colorCenter[index];
			const Color& rColorRim=m_colorRim[index];
			if(rColorCenter[3]<=0.0 && rColorRim[3]<=0.0)
			{
				continue;
			}

			const SpatialVector point=
					m_spOutputPoint->spatialVector(index);

			//* TODO cached 'moved', which is a search

			drawFollicle(spDrawOverlay,point,rimSize,moved(index),FALSE,
					rColorCenter,rColorRim);
		}

		const SpatialVector textPoint(8,48);
		String text;
		text.sPrintf("%s Select",m_areaSelect? "Area": "Single");
		spDrawOverlay->drawAlignedText(textPoint,text,blue);

		spRA=m_spState->getArray("Add");
		if(spRA.isValid())
		{
#if FE_FCO_OUTPUT_DEBUG
			feLog("FollicleOp::handle add count %d\n",spRA->length());
#endif

			for(int i=0;i<spRA->length();i++)
			{
				Record edit=spRA->getRecord(i);

				const SpatialVector point=m_aLocation(edit);

				drawFollicle(spDrawOverlay,point,rimSize,FALSE,TRUE,blank,grey);
			}
		}

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	if(!access(m_spInputPoint,"Input Surface",e_point,e_position)) return;

	if(!accessOutput(m_spOutputPoint,a_rSignal,e_point,e_position)) return;

	if(!accessOutput(m_spOutputNormal,a_rSignal,e_point,e_normal)) return;

	const BWORD paramChanged=!m_brushed;
	m_brushed=FALSE;

	if(paramChanged)
	{
		if(!access(m_spDriver,"Driver Surface")) return;
	}

	//* TODO use
	const BWORD rebuild=(paramChanged ||
			catalog<bool>("Output Surface","replaced"));

	const String hideGroup=catalog<String>("hideGroup");
	if(!hideGroup.empty())
	{
		if(!accessOutput(m_spHideGroup,a_rSignal,
				e_pointGroup,hideGroup)) return;

		const U32 outputCount=m_spOutputPoint->count();

		sp<RecordArray> spRA=m_spState->getArray("Hide");
		if(spRA.isValid())
		{
			for(int i=0;i<spRA->length();i++)
			{
				const I32 pointIndex=m_aIndex(spRA,i);
				if(pointIndex>=0 && pointIndex<outputCount)
				{
					m_spHideGroup->append(pointIndex);
				}
			}
		}

	}

	sp<RecordArray> spRA=m_spState->getArray("Move");
	if(spRA.isValid())
	{
#if FE_FCO_OUTPUT_DEBUG
		feLog("FollicleOp::handle move count %d\n",spRA->length());
#endif
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);

			Record edit=spRA->getRecord(i);

#if FE_FCO_OUTPUT_DEBUG
			feLog("FollicleOp::handle move %d/%d prim %d/%d\n",
					i,spRA->length(),pointIndex,m_spOutputPoint->count());
#endif

			const SpatialVector inputPoint= m_spInputPoint->spatialVector(pointIndex);

			const SpatialVector outputPoint=inputPoint+m_aLocation(edit);

			m_spOutputPoint->set(pointIndex,outputPoint);
		}
	}

	sp<DrawI> spOutputDrawI;
	if(!accessDraw(spOutputDrawI,a_rSignal)) return;

	spRA=m_spState->getArray("Add");
	if(spRA.isValid())
	{
#if FE_FCO_OUTPUT_DEBUG
		feLog("FollicleOp::handle append count %d\n",spRA->length());
#endif

		for(int i=0;i<spRA->length();i++)
		{
			Record edit=spRA->getRecord(i);

#if FE_FCO_OUTPUT_DEBUG
			const I32 pointIndex=m_aIndex(spRA,i);
			feLog("FollicleOp::handle append %d/%d prim %d/%d\n",
					i,spRA->length(),pointIndex,m_spOutputPoint->count());
#endif

			const SpatialVector point=m_aLocation(edit);

			spOutputDrawI->drawPoints(&point,NULL,1,FALSE,NULL);
		}
	}

	return;
}

BWORD FollicleOp::pickToggle(I32 a_index)
{
	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_index;
		if(a_index==pointIndex)
		{
			m_pickArray.erase(m_pickArray.begin()+pickIndex);
			return TRUE;
		}
	}
	pick(a_index);
	return FALSE;
}

void FollicleOp::pick(I32 a_index,Real a_weight)
{
	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_index;
		if(a_index==pointIndex)
		{
			if(rPick.m_weight<a_weight)
			{
				rPick.m_weight=a_weight;
			}
			return;
		}
	}

	m_pickArray.push_back(Pick(a_index,a_weight));
}

Real FollicleOp::picked(I32 a_index)
{
	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_index;
		if(a_index==pointIndex)
		{
			return rPick.m_weight;
		}
	}
	return 0.0;
}

BWORD FollicleOp::hideToggle(I32 a_index)
{
	sp<RecordArray> spRA=m_spState->getArray("Hide");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				spRA->remove(i);
				return TRUE;
			}
		}
	}

	hide(a_index);
	return FALSE;
}

void FollicleOp::hide(I32 a_index)
{
	sp<RecordArray> spRA=m_spState->getArray("Hide");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				return;
			}
		}
	}
	Record hide=m_spScope->createRecord("Hide");
	m_aIndex(hide)=a_index;

	m_spState->add(hide);
}

BWORD FollicleOp::hidden(I32 a_index)
{
	sp<RecordArray> spRA=m_spState->getArray("Hide");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				return TRUE;
			}
		}
	}
	return FALSE;
}

BWORD FollicleOp::moved(I32 a_index)
{
	sp<RecordArray> spRA=m_spState->getArray("Move");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				return TRUE;
			}
		}
	}
	return FALSE;
}

Record FollicleOp::getEdit(I32 a_index)
{
	Record edit=m_spEditMap->lookup(a_index);
	if(!edit.isValid())
	{
		edit=m_spScope->createRecord("Move");
		m_aIndex(edit)=a_index;

		m_spState->add(edit);
	}

	return edit;
}

void FollicleOp::addFollicle(const SpatialVector& a_intersection)
{
	U32 maxIndex=m_spOutputPoint->count();

	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_index;
		if(maxIndex<pointIndex)
		{
			maxIndex=pointIndex;
		}
	}

	Record edit=m_spScope->createRecord("Add");
	m_aIndex(edit)=maxIndex;
	m_aLocation(edit)=a_intersection;

	m_spState->add(edit);
}

void FollicleOp::movePicked(const SpatialVector& a_rDelta)
{
#if FE_FCO_DEBUG
	feLog("FollicleOp::movePicked %s\n",c_print(a_rDelta));
#endif

	const BWORD useUV=catalog<bool>("useUV");

	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_index;
		const Real pointWeight=rPick.m_weight;
		Record edit=getEdit(pointIndex);
		if(useUV)
		{
			slideOnUV(pointIndex,a_rDelta);
		}
		else
		{
			m_aLocation(edit)+=a_rDelta*pointWeight;

#if FE_FCO_DEBUG
			SpatialVector inputPoint;
			set(inputPoint);
			const U32 inputCount=m_spInputPoint->count();
			if(pointIndex<inputCount)
			{
				inputPoint=m_spInputPoint->spatialVector(pointIndex);
			}

			const SpatialVector outputPoint=inputPoint+m_aLocation(edit);

			feLog("FollicleOp::movePicked %d %d at %s by %s to %s\n",
					pickIndex,pointIndex,c_print(inputPoint),
					c_print(m_aLocation(edit)),c_print(outputPoint));
#endif

			snapToSurface(pointIndex);

#if FE_FCO_DEBUG
			feLog("  snap %s\n",c_print(m_aLocation(edit)));
#endif
		}
	}
}

void FollicleOp::slideOnUV(I32 a_index,const Vector2 a_deltaUV)
{
//	feLog("FollicleOp::slideOnUV %d %s\n",a_index,c_print(a_deltaUV));

	const U32 inputCount=m_spInputPoint->count();

	SpatialVector point;
	set(point);
	if(a_index<inputCount)
	{
		point=m_spInputPoint->spatialVector(a_index);
	}

	Record edit=getEdit(a_index);

	sp<SurfaceI::ImpactI> spImpact=m_spDriver->nearestPoint(point);
	if(!spImpact.isValid())
	{
		return;
	}
	const Vector2 origUV=spImpact->uv();

	m_aUV(edit)+=a_deltaUV;
	Vector2 uv=origUV+m_aUV(edit);

	const SpatialVector sample=m_spDriver->samplePoint(uv);

//	feLog("  orig %s add %s uv %s sample %s\n",
//			c_print(origUV),c_print(m_aUV(edit)),c_print(uv),c_print(sample));

	m_aLocation(edit)=sample-point;
}

void FollicleOp::snapToSurface(I32 a_index)
{
	const U32 inputCount=m_spInputPoint->count();

	SpatialVector point;
	set(point);
	if(a_index<inputCount)
	{
		point=m_spInputPoint->spatialVector(a_index);
	}

	Record edit=getEdit(a_index);

	sp<SurfaceI::ImpactI> spImpact=
			m_spDriver->nearestPoint(point+m_aLocation(edit));
	if(!spImpact.isValid())
	{
		return;
	}
	const SpatialVector intersection=spImpact->intersection();

	m_aLocation(edit)=intersection-point;
}

void FollicleOp::drawFollicle(sp<DrawI>& a_rspDraw,
	const SpatialVector& a_rPoint,Real a_rimSize,
	BWORD a_moved,BWORD a_added,
	const Color& a_rColorCenter,const Color& a_rColorRim)
{
	sp<ViewI> spView=a_rspDraw->view();

	const Vector2i projected=spView->project(a_rPoint,ViewMode::e_perspective);
	const Vector2i rim(a_rimSize,a_rimSize);

	Box2i box=projected-rim;
	box.size()=2.0*rim+Vector2i(1,1);

	if(a_rColorCenter[3]>0.0)
	{
		const I32 lineWidth=catalog<I32>("lineWidth");

		SpatialVector corner[2];
		corner[0]=SpatialVector(box)+SpatialVector(lineWidth,lineWidth);
		corner[1]=SpatialVector(box)-SpatialVector(lineWidth,lineWidth)+
				SpatialVector(box.size());

		a_rspDraw->drawRectangles(corner,2,FALSE,&a_rColorCenter);
	}

	a_rspDraw->drawBox(box,a_rColorRim);

	if(a_moved)
	{
		Color shadow(0.4,1.0,1.0);
		SpatialVector glint=SpatialVector(box)+
				SpatialVector(box.size()[0],0.0,1.0);
		a_rspDraw->drawPoints(&glint,NULL,1,FALSE,&shadow);
	}

	if(a_added)
	{
		Color sparkle(1.0,1.0,0.6);
		SpatialVector glint=SpatialVector(box)+
				SpatialVector(0.0,box.size()[1],1.0);
		a_rspDraw->drawPoints(&glint,NULL,1,FALSE,&sparkle);
	}
}

void FollicleOp::setupState(void)
{
	m_aIndex.populate(m_spScope,"Move","Index");
	m_aLocation.populate(m_spScope,"Move","Location");
	m_aUV.populate(m_spScope,"Move","UV");

	m_aIndex.populate(m_spScope,"Add","Index");
	m_aLocation.populate(m_spScope,"Add","Location");
	m_aUV.populate(m_spScope,"Add","UV");

	m_aIndex.populate(m_spScope,"Hide","Index");

	m_spEditMap=m_spState->lookupMap(m_aIndex);
}

BWORD FollicleOp::loadState(const String& rBuffer)
{
	BWORD success=OperatorState::loadState(rBuffer);

	m_spEditMap=m_spState->lookupMap(m_aIndex);

	return success;
}
